(* Copyright (c) 2021 The Proofgold Lava developers *)
(* Copyright (c) 2020 The Proofgold developers *)
(* Copyright (c) 2016 The Qeditas developers *)
(* Copyright (c) 2017-2018 The Dalilcoin developers *)
(* Distributed under the MIT software license, see the accompanying
   file COPYING or http://www.opensource.org/licenses/mit-license.php. *)

open Ser
open Hashaux
open Hash
open Logic
open Mathdata


let counterbound = 150000000
exception CheckingBd
let cnt counter = (incr counter; if false (* !counter >= counterbound *) then raise CheckingBd)
let dbbound = 256

(* <<<<<<<<<<<<<<<<<<<<<<<< *)

let ty_no_h=(Hashtbl.create 1000 : (stp, int) Hashtbl.t);;
let no_ty_h=(Hashtbl.create 1000 : (int, stp) Hashtbl.t);;
let ty_num=ref 0;;

let ty_f ty =
  try Hashtbl.find ty_no_h ty
  with Not_found ->
    incr ty_num;
    Hashtbl.add ty_no_h ty !ty_num;
    Hashtbl.add no_ty_h !ty_num ty;
    !ty_num
and ty_f_rev = Hashtbl.find no_ty_h;;

let hv_no=(Hashtbl.create 1000 : (hashval, int) Hashtbl.t);;
let no_hv=(Hashtbl.create 1000 : (int, hashval) Hashtbl.t);;
let hh_num=ref 0;;

let hh_no s =
  try Hashtbl.find hv_no s
  with Not_found ->
    incr hh_num;
    Hashtbl.add hv_no s !hh_num;
    Hashtbl.add no_hv !hh_num s;
    !hh_num
and hh_no_rev = Hashtbl.find no_hv;;

external ftrm_hash_clear : unit -> unit = "c_hash_clear" [@@noalloc];;

let ftrm_hash_clear () =
  ftrm_hash_clear ();
  Hashtbl.reset ty_no_h;
  Hashtbl.reset no_ty_h;
  Hashtbl.reset hv_no;
  Hashtbl.reset no_hv;
  ty_num := 0;
  hh_num := 0;;

type ftp=int;;
type ftm=int;;

external init_dbs : unit -> unit = "c_init_dbs";;
init_dbs ();;

external mk_db : int -> ftm = "c_mk_db" [@@noalloc];;
let mk_db i =
  if i >= dbbound then (Printf.printf "mk_db too high: %d\n" i; flush stdout; raise CheckingBd);
  mk_db i;;
external mk_tmh_internal : int -> ftm = "c_mk_tmh" [@@noalloc];;
let mk_tmh s = mk_tmh_internal (hh_no s);;
external mk_prim : int -> ftm = "c_mk_prim" [@@noalloc];;
external mk_ap : ftm -> ftm -> ftm = "c_mk_norm_ap" [@@noalloc];;
external mk_imp : ftm -> ftm -> ftm = "c_mk_imp" [@@noalloc];;
external mk_lam : ftp -> ftm -> ftm = "c_mk_norm_lam" [@@noalloc];;
external mk_all : ftp -> ftm -> ftm = "c_mk_all" [@@noalloc];;

type ftm_tag =
| FT_None
| FT_DB
| FT_TmH
| FT_Prim
| FT_Ap
| FT_Lam
| FT_Imp
| FT_All;;

external get_tag : ftm -> ftm_tag = "c_get_tag" [@@noalloc];;

external get_no : ftm -> int = "c_get_no" [@@noalloc];;
external get_l : ftm -> ftm = "c_get_l" [@@noalloc];;
external get_r : ftm -> ftm = "c_get_r" [@@noalloc];;
external get_maxv : ftm -> int = "c_get_maxv" [@@noalloc];;
external get_fv0_0 : ftm -> bool = "c_get_fv0_0" [@@noalloc];;

external uptrm : ftm -> int -> int -> ftm = "c_uptrm" [@@noalloc];;
external subst : ftm -> int -> ftm -> ftm = "c_subst" [@@noalloc];;

external print : ftm -> unit = "c_print" [@@noalloc];;
external size : ftm -> int = "c_size" [@@noalloc];;

let mk_ex a t1 =
  let t2 = uptrm t1 1 1 in
  mk_all (ty_f Prop) (mk_imp (mk_all a (mk_imp t2 (mk_db 1))) (mk_db 0));;

let mk_eq a t1 t2 =
  let t1b = uptrm t1 0 1 in
  let t2b = uptrm t2 0 1 in
  let at = ty_f_rev a in
  let ty_aap = ty_f (TpArr(at,TpArr(at,Prop))) in
  mk_all ty_aap (mk_imp (mk_ap (mk_ap (mk_db 0) t1b) t2b) (mk_ap (mk_ap (mk_db 0) t2b) t1b));;

let ftrm_trm counter m = failwith "ftrm_trm: unimplemented";;
let ftrma_trm counter (m,_,_,_) = failwith "ftrma_trm: unimplemented";;

let rec trm_ftrma m =
  match m with
  | TmH(h) -> mk_tmh h
  | DB(i) -> mk_db i
  | Prim(i) -> mk_prim i
  | Ap(m1,m2) -> mk_ap (trm_ftrma m1) (trm_ftrma m2)
  | Lam(a,m1) -> mk_lam (ty_f a) (trm_ftrma m1)
  | Imp(m1,m2) -> mk_imp (trm_ftrma m1) (trm_ftrma m2)
  | All(a,m1) -> mk_all (ty_f a) (trm_ftrma m1)
  | Ex(a,m1) -> mk_ex (ty_f a) (trm_ftrma m1)
  | Eq(a,m1,m2) -> mk_eq (ty_f a) (trm_ftrma m1) (trm_ftrma m2)

(* >>>>>>>>>>>>>>>>>>>>>>> *)
(* <<<<<<<<<<<<<<<<<<<<<<< *)

let rec get_stp_trm_raise counter ctx tys t thy =
  cnt counter;
  match t with
  | DB i -> List.nth ctx i
  | TmH h -> Hashtbl.find tys h
  | Prim i -> List.nth thy i
  | Ap (t1, t2) ->
     let s = get_stp_trm_raise counter ctx tys t1 thy in
     (match s with
      | TpArr (b, alpha) ->
         let b2 = get_stp_trm_raise counter ctx tys t2 thy in
         if b2 = b then alpha else raise Not_found
      | _ -> raise Not_found)
  | Lam (a1, t1) ->
     TpArr (a1, get_stp_trm_raise counter (a1 :: ctx) tys t1 thy)
  | Imp (t1, t2) ->
     let a = get_stp_trm_raise counter ctx tys t1 thy in
     let b = get_stp_trm_raise counter ctx tys t2 thy in
     if a = Prop && b = Prop then Prop else raise Not_found
  | All (b, t1) | Ex (b, t1) ->
     if get_stp_trm_raise counter (b :: ctx) tys t1 thy = Prop then Prop else raise Not_found
  | Eq (b, t1, t2) ->
     let b1 = get_stp_trm_raise counter ctx tys t1 thy
     and b2 = get_stp_trm_raise counter ctx tys t2 thy in
     if b1 = b && b2 = b then Prop else raise Not_found;;





let head_args t =
  let rec ha args tt =
    match get_tag tt with
    | FT_Ap -> ha ((get_r tt)::args) (get_l tt)
    | _ -> (tt,args)
  in
  ha [] t;;

let delta defs h args =
  match Hashtbl.find defs (hh_no_rev (get_no h)) with
  | (_, def) -> Some (List.fold_left mk_ap def args)
  | exception Not_found -> None;;

(*
let rec conv defs m n =
  m == n ||
  match (head_args m), (head_args n) with
  | (mh, (_ :: _ as ma)), (nh, (_ :: _ as na)) when
     (mh == nh && List.length ma = List.length na && List.for_all2 (conv defs) ma na) -> true
  | (F_TmH mhd, ma), _ when Hashtbl.mem defs mhd ->
     (match delta defs mhd ma with Some m' -> conv defs m' n | None -> false)
  | _, (F_TmH nhd, na) ->
     (match delta defs nhd na with Some n' -> conv defs m n' | None -> false)
  | (F_Imp(m1,m2), _), (F_Imp(n1,n2), _) -> conv defs m1 n1 && conv defs m2 n2
  | (F_All(a1,m1), _), (F_All(b1,n1), _)
  | (F_Lam(a1,m1), _), (F_Lam(b1,n1), _) -> a1 = b1 && conv defs m1 n1
  | _ -> false;;
*)

(* List version to translate to imperative code. *)
let rec conv defs l =
  match l with
  | [] -> true
  | (m, n) :: t ->
     if m == n then conv defs t else
     let (mh, ma) = head_args m and (nh, na) = head_args n in
     match ma, na with (_ :: _, _ :: _) when
       (mh == nh && List.length ma = List.length na && conv defs (List.combine ma na)) -> conv defs t
     | _ ->
       match get_tag mh, get_tag nh with
     | FT_TmH, _ when Hashtbl.mem defs (hh_no_rev (get_no mh)) ->
        (match delta defs mh ma with Some m' -> conv defs ((m', n) :: t) | None -> false)
     | _, FT_TmH ->
        (match delta defs nh na with Some n' -> conv defs ((m, n') :: t) | None -> false)
     | FT_Imp, FT_Imp -> conv defs ((get_l m, get_l n) :: (get_r m, get_r n) :: t)
     | FT_All, FT_All
     | FT_Lam, FT_Lam -> get_no m = get_no n && conv defs ((get_r m, get_r n) :: t)
     | _ -> false;;

let rec headnorm defs t =
  let th, ta = head_args t in
  match get_tag th with
  | FT_TmH ->
     (match Hashtbl.find defs (hh_no_rev (get_no th)) with
     | (_, def) -> headnorm defs (List.fold_left mk_ap def ta)
     | exception Not_found -> t)
  | _ -> t;;


let rec get_prop_pf_raise counter ctx phi tys defs kl p thy : ftm =
  cnt counter;
  match p with
  | Known h -> snd (Hashtbl.find kl h)
  | Hyp i ->
     let x = List.nth phi i in let (i, t) = !x in
     if i = 0 then t else (let t' = uptrm t 0 i in x := (0, t'); t')
  | PrAp (p1, p2) ->
     let t = headnorm defs (get_prop_pf_raise counter ctx phi tys defs kl p1 thy) in
     if get_tag t <> FT_Imp then raise Not_found;
     let t1 = get_l t and t2 = get_r t in
     if not (conv defs [get_prop_pf_raise counter ctx phi tys defs kl p2 thy, t1]) then raise Not_found;
     t2
  | TmAp (p1, t1) ->
     let t = headnorm defs (get_prop_pf_raise counter ctx phi tys defs kl p1 thy) in
     if get_tag t <> FT_All then raise Not_found;
     let a = get_no t and m = get_r t in
     let cx = get_maxv m in
     if ty_f (get_stp_trm_raise counter ctx tys t1 thy) <> a then raise Not_found;
     if cx <= 0 then m else
     if get_fv0_0 m then uptrm m 0 (-1) else
     let t1b = trm_ftrma t1 in
     subst m 0 t1b
  | PrLa (s, p1) ->
     if get_stp_trm_raise counter ctx tys s thy <> Prop then raise Not_found;
     let q = trm_ftrma s in
     mk_imp q (get_prop_pf_raise counter ctx ((ref (0, q)) :: phi) tys defs kl p1 thy)
  | TmLa (a1, p1) ->
     let phi2 = List.map (fun x -> let (i, t) = !x in ref (i + 1, t)) phi in
     mk_all (ty_f a1) (get_prop_pf_raise counter (a1 :: ctx) phi2 tys defs kl p1 thy)
  | Ext (a, b) ->
     let ta = ty_f a and tb = ty_f b and taab = ty_f (TpArr(a,b)) in
     mk_all taab (mk_all taab (mk_imp
       (mk_all ta (mk_eq tb (mk_ap (mk_db 2) (mk_db 0)) (mk_ap (mk_db 1) (mk_db 0))))
       (mk_eq taab (mk_db 1) (mk_db 0))))

(** val correct_pf_f : int ref ->
    stp list -> trm list -> gsign -> pf -> trm -> stp list -> bool **)
let correct_pf_f counter ctx phi tys defs kl p t thy =
(*  match (p,t) with
  | (TmLa(a1,p1),(F_All(a2,t2),_,_,_)) -> (** assume the a2 = a1, to test savings of making a1 optional here **)
     correct_pf_f counter (ty_f_rev a2 :: ctx) (List.map (fun x -> let (i, t) = !x in ref (i + 1, t)) phi) tys defs kl p1 t2 thy
  | (PrLa(s, p1),(F_Imp(q1, q2),_,_,_)) -> (** do not check the s converts to q1, to test savings of making s optional here **)
     correct_pf_f counter ctx ((ref (0, q1)) :: phi) tys defs kl p1 q2 thy
  | _ ->*)
     match get_prop_pf_raise counter ctx phi tys defs kl p thy with
     | pp -> conv defs [pp, t]
     | exception Not_found -> false (* List.find *)
     | exception Failure _ -> false (* List.nth *)

(*
let hashroots = Array.make 10000000 None;;

let rec tm_hashroot_fun sf (id : int) m =
  match hashroots.(id) with
  | Some r -> r
  | None ->
    let ret =
    match sf,m with
    | _, TmH(h) -> h
    | _, Prim(x) -> hashtag (hashint32 (Int32.of_int x)) 96l
    | _, DB(x) -> hashtag (hashint32 (Int32.of_int x)) 97l
    | [hm; hn], Ap(m,n) -> hashtag (hashpair hm hn) 98l
    | [hm; hn], Imp(m,n) -> hashtag (hashpair hm hn) 100l
    | [hm], Lam(a,m) -> hashtag (hashpair (hashtp a) hm) 99l
    | [hm], All(a,m) -> hashtag (hashpair (hashtp a) hm) 101l
    | [hm], Ex(a,m) -> hashtag (hashpair (hashtp a) hm) 102l
    | [hm; hn], Eq(a,m,n) -> hashtag (hashpair (hashpair (hashtp a) hm) hn) 103l
    | _ -> failwith "tm_hashroot_fun"
    in
    hashroots.(id) <- Some ret;
    ret;;

let tm_hashroot t = Utm.fold_id (Obj.magic tm_hashroot_fun) (Obj.magic t);;
*)

(** val free_trm_trm : int ref -> trm -> int -> bool **)

let rec free_trm_trm counter t i =
  cnt counter;
  match t with
  | DB j -> i = j
  | Ap (m1, m2) -> (||) (free_trm_trm counter m1 i) (free_trm_trm counter m2 i)
  | Lam (_, m1) -> free_trm_trm counter m1 ((+) i 1)
  | Imp (m1, m2) -> (||) (free_trm_trm counter m1 i) (free_trm_trm counter m2 i)
  | All (_, m1) -> free_trm_trm counter m1 ((+) i 1)
  | Ex (_, m1) -> free_trm_trm counter m1 ((+) i 1)
  | Eq (_, m1, m2) -> (||) (free_trm_trm counter m1 i) (free_trm_trm counter m2 i)
  | _ -> false

(** val is_norm : int ref -> trm -> bool **)

let rec is_norm counter m =
  cnt counter;
  match m with
  | Ap (Lam (_, _), m2) -> false
  | Ap (m1, m2) -> is_norm counter m1 && is_norm counter m2
  | Lam (_, Ap (f, DB 0)) when not (free_trm_trm counter f 0) -> false
  | Lam (_, m1) -> is_norm counter m1
  | Imp (m1, m2) -> is_norm counter m1 && is_norm counter m2
  | All (_, m1) -> is_norm counter m1
  | Ex (_, _) -> false
  | Eq (_, _, _) -> false
  | _ -> true

let rec tm_tp_f gvtp tys th h a =
  try let b = Hashtbl.find tys h in a = b
  with Not_found -> gvtp th h a;;

(** val check_doc_f : int ref ->
    (hashval option -> hashval -> stp -> bool) -> (hashval option -> hashval
    -> bool) -> hashval option -> theory -> stree option -> doc ->
    (gsign_f * hashval list) option
 where gsign_f is
  ((hashval * stp) * trm option * ftrma option) list * (hashval * trm * ftrma) list
 **)

let rec check_doc_f counter gvtp gvkn th thy d =
  let tys = Hashtbl.create 100 and defs = Hashtbl.create 100 and kl = Hashtbl.create 100 in
  let check_doc_f1 d0 () =
    match d0 with
    | Docparam (h, a) ->
       if not (tm_tp_f gvtp tys th h a) then raise Exit else
       Hashtbl.add tys h a
    | Docdef (_, TmH _) -> ()
    | Docdef (a, m) ->
       if not (is_norm counter m) then raise Exit else
       let h = tm_hashroot m in
       let m2 = trm_ftrma m in
       Hashtbl.add tys h a;
       Hashtbl.add defs h (m, m2)
    | Docknown p ->
       if not (is_norm counter p) then raise Exit else
       if get_stp_trm_raise counter [] tys p (fst thy) <> Prop then raise Exit else
       let k = tm_hashroot p in
       let exists_fun x = (x = k) in
       if not (List.exists exists_fun (snd thy)) &&
          not (Hashtbl.mem kl k || gvkn th k) then raise Exit else
       let p2 = trm_ftrma p in
       Hashtbl.add kl k (p, p2)
    | Docpfof (p, d) ->
       if not (is_norm counter p) then raise Exit;
       if get_stp_trm_raise counter [] tys p (fst thy) <> Prop then raise Exit;
       let p2 = trm_ftrma p in
       if not (correct_pf_f counter [] [] tys defs kl d p2 (fst thy)) then raise Exit else
       let k = tm_hashroot p in
       Hashtbl.add kl k (p, p2)
    | Docconj p ->
       if not (is_norm counter p) then raise Exit else
       if get_stp_trm_raise counter [] tys p (fst thy) <> Prop then raise Exit else ()
    | Docsigna h -> raise (Failure "signatures are not allowed") (* C says "it's questionable if it even works" *)
  in
  try
    List.fold_right check_doc_f1 d ();
    let fdef_to_def h =
      match Hashtbl.find defs h with
      | (d, _) -> Some d
      | exception Not_found -> None in
    let tmtpl = Hashtbl.fold (fun h ty sf -> ((h, ty), fdef_to_def h) :: sf) tys [] in
    let kl = Hashtbl.fold (fun k (p, _) sf -> (k, p) :: sf) kl [] in
    Some (tmtpl, kl)
  with Exit -> None;;

let check_doc counter gvtp gvkn th thy d =
  ftrm_hash_clear();
  (** must be cleared before checking docs or can get mixed with transparent defs from previous docs; must not be cleared within docs since the intermediate signatures have ftrms **)
  check_doc_f counter gvtp gvkn th thy d;;
